from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest
import rasterio as rio
import scipy

from mnp.species_models.environmental_factor import EnvironmentalFactor


@pytest.fixture
def parameters(request):
    mnp_parameters = Mock()
    mnp_parameters.geospatial_profile = Mock()
    mnp_parameters.folders = {"environmental_factor": "."}
    mnp_parameters.environmental_response_tables = {
        "ndep": pd.read_csv(
            request.path.parent / "data/parameters/ndep_suitability_index.csv",
            index_col=0,
        ),
        "gvg": pd.read_csv(
            request.path.parent / "data/parameters/gvg_suitability_index.csv",
            index_col=0,
        ),
    }
    mnp_parameters.species_traits = pd.read_csv(
        request.path.parent / "data/parameters/species_traits.csv", index_col=0
    )
    mnp_parameters.response_020 = 0
    mnp_parameters.response_080 = 0.5
    mnp_parameters.response_100 = 1
    return mnp_parameters


@pytest.fixture
def expected_environmental_factor(request):
    with rio.open(
        request.path.parent / "data/rasters/environmental_factor_60002.tif"
    ) as src:
        return scipy.sparse.csr_array(src.read(1))


@pytest.fixture
def spatial_data_coverage(request):
    with rio.open(
        request.path.parent / "data/rasters/spatial_data_coverage.tif"
    ) as src:
        return scipy.sparse.csr_array(src.read(1))


@pytest.fixture
def environmentals(request, spatial_data_coverage):
    env_dict = {}
    for environmental_name in ["ndep", "gvg"]:
        with rio.open(
            request.path.parent
            / f"data/rasters/environmentals/{environmental_name}.tif"
        ) as src:
            env_dict[environmental_name] = spatial_data_coverage * src.read(1)
    return env_dict


class TestEnvironmentalFactor:
    def test_calculate(self, parameters, environmentals, expected_environmental_factor):
        environmental_factor = EnvironmentalFactor(parameters, 60002)
        environmental_factor.calculate(environmentals)
        assert environmental_factor.array.max() > 0
        np.testing.assert_allclose(
            environmental_factor.array.todense(),
            expected_environmental_factor.todense(),
        )
