import pytest
import rasterio as rio
import scipy.sparse

from mnp.species_models import habitat_suitability


@pytest.fixture
def land_types(request):
    return {
        path.stem: scipy.sparse.load_npz(path)
        for path in (request.path.parent / "data/rasters/land_types_aggregated").glob(
            "*.npz"
        )
    }


@pytest.fixture
def environmental_factor(request):
    with rio.open(
        request.path.parent / "data/rasters/environmental_factor_60002.tif"
    ) as src:
        return scipy.sparse.csr_array(src.read(1))


# setup Nature HSI procedure
# setup land types
class TestNature:
    def test_hsi_with_no_matching_land_types(self, land_types):
        hsi_procedure = habitat_suitability.Nature(0.1, {"nonexisting": 1.0})
        hsi = hsi_procedure.run(land_types, None)
        assert hsi.min() == 0
        assert hsi.max() == 0

    def test_hsi_with_matching_land_types(self, land_types):
        hsi_procedure = habitat_suitability.Nature(
            0.1, {"N02.00.00": 0.5, "N03.00.00": 0.3}
        )
        hsi = hsi_procedure.run(land_types, None)
        assert hsi.max() == 0.5
        assert hsi.min() >= 0

    def test_hsi_with_high_threshold(self, land_types):
        hsi_procedure = habitat_suitability.Nature(
            0.6, {"N02.00.00": 0.5, "N03.00.00": 0.3}
        )
        hsi = hsi_procedure.run(land_types, None)
        assert hsi.max() == 0.0
        assert hsi.min() == 0.0

    def test_hsi_with_environmental_factor(self, land_types, environmental_factor):
        hsi_procedure = habitat_suitability.Nature(
            0.1, {"N02.00.00": 0.5, "N03.00.00": 0.3}
        )
        hsi = hsi_procedure.run(land_types, environmental_factor)
        assert hsi.max() <= (0.5 * environmental_factor.max())
        assert hsi.min() >= 0.0

    def test_hsi_with_environmental_factor_under_threshold(
        self, land_types, environmental_factor
    ):
        hsi_procedure = habitat_suitability.Nature(
            0.1, {"N02.00.00": 0.5, "N03.00.00": 0.3}
        )
        hsi = hsi_procedure.run(land_types, environmental_factor * 0.05)
        assert hsi.max() == 0
        assert hsi.min() == 0
