from unittest.mock import Mock

import pandas as pd
import pytest
from scipy.sparse import csr_array

from mnp.species_models.species_evaluation import SpeciesEvaluation


@pytest.fixture
def parameters(request):
    mnp_parameters = Mock()
    mnp_parameters.geospatial_profile = Mock()
    mnp_parameters.folders = {"environmental_factor": "."}
    mnp_parameters.species_traits = pd.DataFrame(
        columns=[
            "species_code",
            "local_distance",
            "key_population_area",
            "group_traits_code",
            "environmental_factor_formula",
            "land_type_procedure",
        ],
        data=[[60002, 2, 3, "E", "T*gvg*ph*ndep", "Nature"]],
    ).set_index("species_code")
    mnp_parameters.group_traits = pd.read_csv(
        request.path.parent / "data/parameters/group_traits.csv", index_col=0
    )
    mnp_parameters.pxl_area = 10000
    mnp_parameters.small_pop_slope = 2
    mnp_parameters.small_pop_threshold_area = 500
    return mnp_parameters


@pytest.fixture
def hsi():
    hsi_mock = Mock()
    hsi_mock.array = csr_array(
        [
            [0, 0, 1, 0.8, 0.5],
            [0.5, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0.5, 0, 0, 0],
        ]
    )
    return hsi_mock


@pytest.fixture
def clustering():
    clustering_mock = Mock()
    clustering_mock.array = csr_array(
        [
            [0, 0, 1, 1, 1],
            [1, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 2, 0, 0, 0],
        ]
    )
    return clustering_mock


@pytest.fixture
def species_evaluation(parameters, hsi, clustering):
    return SpeciesEvaluation(parameters, 60002, hsi, clustering)


class TestSpeciesEvaluation:

    def test_species_evaluation_results(self, species_evaluation):
        species_evaluation.calculate()
        results = species_evaluation.results()
        assert results["populations"] == 2
        assert results["populations_key_population"] == 1
        assert results["total_area_ha"] == 6
        assert results["total_effective_area_ha"] == 4.3
        assert results["total_effective_area_kp"] == 4.3 / 3
        assert results["population_is_key_population"] == {1: True, 2: False}
